import glob
import json
import os
from datetime import datetime
from test.utils.api_client import GitHubAPIClient
from test.utils.config import GitHubCIConfig
from test.utils.data import (
    PerformanceComparisonMetadata,
    PerformanceData,
    PerformanceMetadata,
    PerformanceMetricComparison,
    calculate_quantiles,
)
from test.utils.files import (
    cleanup_output_dir,
    cleanup_zip_file,
    download_zip_file,
    extract_json_from_zip,
    find_performance_artifact,
    load_json_file,
    save_json_file,
)
from typing import Dict, Optional

import pandas as pd
import pytest
from rich.console import Console
from rich.table import Table

from vajra.benchmark.benchmark_runner import BenchmarkRunnerLauncher
from vajra.benchmark.config import (
    BaseRequestIntervalGeneratorConfig,
    BenchmarkConfig,
    SyntheticRequestGeneratorConfig,
    TraceRequestLengthGeneratorConfig,
)
from vajra.config import (
    FcfsFixedChunkSchedulerConfig,
    MetricsConfig,
    ModelConfig,
    ParallelConfig,
)
from vajra.config.flat_dataclass import get_config_class_by_type_name
from vajra.logger import init_logger
from vajra.types import RequestIntervalGeneratorType

console: Console = Console()  # Rich console for colored output
logger = init_logger(__name__)


def run_perf_test_benchmark(
    model: str,
    max_model_len: int,
    pp_size: int,
    tp_size: int,
    dataset: str,
    request_pattern: str,
    benchmark_output_dir: str,
    test_output_dir: str,
    qps: int,
    config: GitHubCIConfig,
) -> None:
    """
    Configures and runs a single performance test.

    Args:
        model: The name of the model.
        max_model_len: The maximum model length.
        pp_size: The pipeline parallel size.
        tp_size: The tensor parallel size.
        dataset: The dataset to use.
        request_pattern: The request pattern to use.
        benchmark_output_dir: Directory for benchmark output.
        test_output_dir: Directory for test output.
        qps: Queries per second.
        config: CI configuration class
    """

    model_config = ModelConfig(model=model, max_model_len=max_model_len)
    parallel_config = ParallelConfig(
        pipeline_parallel_size=pp_size, tensor_parallel_size=tp_size
    )

    request_length_generator_config = TraceRequestLengthGeneratorConfig(
        trace_file=f"data/processed_traces/{dataset}.csv"
    )

    request_interval_generator_config = get_config_class_by_type_name(
        BaseRequestIntervalGeneratorConfig,
        RequestIntervalGeneratorType(request_pattern),
    )(**{"qps": qps})

    request_generator_config = SyntheticRequestGeneratorConfig(
        length_generator_config=request_length_generator_config,
        interval_generator_config=request_interval_generator_config,
    )

    scheduler_config = FcfsFixedChunkSchedulerConfig(chunk_size=512)

    metrics_config = MetricsConfig(
        write_metrics=True,
        enable_chrome_trace=True,
    )

    benchmark_config = BenchmarkConfig(
        log_level="error",
        model_config=model_config,
        output_dir=benchmark_output_dir,
        parallel_config=parallel_config,
        request_generator_config=request_generator_config,
        scheduler_config=scheduler_config,
        metrics_config=metrics_config,
    )

    # TODO(chus) performance output metrics as dataclass
    benchmark_runner = BenchmarkRunnerLauncher(benchmark_config)
    benchmark_runner.run()

    key = f"{model.replace('/', '-')}_pp{pp_size}_tp{tp_size}_{dataset}_{request_pattern}_{qps}qps"

    perf_json_path = os.path.join(
        test_output_dir, f"perf_results_{os.getenv('GITHUB_RUN_ID', 'dev')}.json"
    )

    perf_data = _build_perf_data(benchmark_output_dir)
    _update_json_file(perf_json_path, key, perf_data, config)
    cleanup_output_dir(benchmark_output_dir)


@pytest.fixture(scope="session")
def performance_comparison_session() -> (
    Optional[Dict]
):  # Return type hint for baseline_metrics
    """
    Session-scoped fixture to handle performance comparison logic in CI,
    or just run performance benchmarks locally.
    Returns baseline metrics if available, otherwise None.
    """
    config = GitHubCIConfig()
    baseline_metrics = None
    most_recent_run_id = None
    api_client = GitHubAPIClient(
        config.github_token, config.repo_owner, config.repo_name
    )

    try:
        # fetch most recent successful workflow run (once per session)
        workflow_data = api_client.get_workflow_runs(
            config.workflow_filename, config.baseline_branch
        )
        if not workflow_data["total_count"]:
            logger.warning(
                "No workflow runs found for baseline branch. Skipping performance comparison."
            )
        else:
            for workflow in workflow_data["workflow_runs"]:
                if workflow["name"] == config.workflow_name:
                    most_recent_run_id = workflow["id"]
                    break

            if not most_recent_run_id:
                logger.warning(
                    f"No successful workflow runs with name '{config.workflow_name}' found for baseline. Skipping performance comparison."
                )
            else:
                logger.info(f"Baseline workflow run ID: {most_recent_run_id}")

                # get artifacts for the previous workflow run (once per session)
                artifacts_data = api_client.get_artifacts_for_run(most_recent_run_id)
                found_artifact = find_performance_artifact(
                    artifacts_data,
                    most_recent_run_id,
                    config.performance_artifact_prefix,
                )

                if found_artifact:
                    # download and extract baseline artifact (once per session)
                    artifact_zip_path = download_zip_file(
                        api_client,
                        found_artifact["id"],
                        config.output_dir,
                        found_artifact["name"],
                    )
                    prev_json_files = extract_json_from_zip(
                        artifact_zip_path, config.output_dir
                    )
                    cleanup_zip_file(artifact_zip_path)

                    if not prev_json_files:
                        logger.warning("No JSON file found in baseline artifact.")
                    else:
                        prev_json_path = prev_json_files[0]
                        baseline_metrics = load_json_file(prev_json_path)
                        logger.info(f"Baseline metrics loaded.")
                else:
                    logger.info("No baseline performance artifact found.")

    except Exception as e:
        console.print_exception(show_locals=True)
        logger.error(f"Error during baseline artifact retrieval: {e}", exc_info=True)
        baseline_metrics = None

    yield baseline_metrics

    # compare and save results (after all tests in session)
    if baseline_metrics:
        try:
            # load performance from this run
            new_json_path = os.path.join(
                config.output_dir, f"perf_results_{config.run_id}.json"
            )
            new_metrics = load_json_file(new_json_path)
            logger.info(f"Loaded new metrics for comparison.")

            comparison_results_data = compare_metrics(
                baseline_metrics, new_metrics, config.regression_threshold_pct
            )

            perf_regression = report_performance_comparison(
                comparison_results_data, config.regression_threshold_pct
            )

            # dataclass to dictionary for JSON serialization
            comparison_results = {
                model_name: {
                    metric_name: metrics.__dict__
                    for metric_name, metrics in model_comparison.items()
                }
                for model_name, model_comparison in comparison_results_data.items()
            }
            # save comparison results
            metadata = create_comparison_metadata(config, most_recent_run_id)
            comparison_output = {"results": comparison_results, "metadata": metadata}
            save_comparison_results(
                comparison_output=comparison_output,
                output_dir=config.output_dir,
                run_id=config.run_id,
                filename_prefix=config.comparison_result_prefix,
            )

            if perf_regression:
                logger.error("Performance regression detected!")
                pytest.fail(
                    "Performance regression detected in one or more metrics. Check output for details."
                )

        except Exception as e:
            console.print_exception(show_locals=True)
            logger.error(f"Error during performance comparison and result saving: {e}")
    elif baseline_metrics is None:
        logger.error(
            "Skipped performance comparison due to missing baseline metrics (or errors during retrieval)."
        )
        pytest.fail(
            f"Could not find baseline metrics or successful test suite run for workflow '{config.workflow_name}' in "
            f"branch '{config.baseline_branch}'."
        )


@pytest.mark.parametrize(
    "model, max_model_len, pp_size, tp_size, dataset, request_pattern, qps",
    [
        (
            "meta-llama/Meta-Llama-3-8B",
            4096,
            1,
            1,
            "sharegpt_8k_stats_llama2_tokenizer_filtered_v2",
            "POISSON",
            100,
        ),
    ],
)
@pytest.mark.performance
def test_perf_benchmark(
    performance_comparison_session: Optional[
        Dict
    ],  # access baseline_metrics via fixture
    model: str,
    max_model_len: int,
    pp_size: int,
    tp_size: int,
    dataset: str,
    request_pattern: str,
    qps: int,
) -> None:
    """
    Entry point for parameterized performance benchmarks.
    The performance comparison is handled by the session-scoped fixture 'performance_comparison_session'.
    """
    config = GitHubCIConfig()  # Get config - needed for output dir
    benchmark_output_dir = os.path.join(os.getcwd(), "benchmark_output")
    test_output_dir = config.output_dir

    try:
        run_perf_test_benchmark(
            model,
            max_model_len,
            pp_size,
            tp_size,
            dataset,
            request_pattern,
            benchmark_output_dir,
            test_output_dir,
            qps,
            config,
        )  # run benchmark for this parameter set

    except (
        Exception
    ) as e:  # catch exceptions from benchmark run only, fixture handles comparison errors
        console.print_exception(show_locals=True)
        logger.error(f"Performance benchmark run error: {e}", exc_info=True)
        pytest.fail(
            f"Performance benchmark run failed for parameters: model={model}, max_model_len={max_model_len}, pp_size={pp_size}, tp_size={tp_size}, dataset={dataset}, request_pattern={request_pattern}"
        )


def _build_perf_data(benchmark_output_dir: str) -> PerformanceData:
    """
    Builds performance data from benchmark output and updates a JSON file.

    Args:
        benchmark_output_dir: The directory containing benchmark output.
    """

    metrics = {
        "ttft": ("prefill_e2e_time", "prefill_e2e_time"),
        "tbt": (
            "decode_token_execution_plus_preemption_time",
            "decode_token_execution_plus_preemption_time",
        ),
    }

    perf_data_dict = {}
    try:
        # assumption: 1 replica and 1 benchmark output
        csv_main_path = glob.glob(f"{benchmark_output_dir}/*/replica_0/plots/")[0]
        for prefix, (file_name, col_name) in metrics.items():
            csv_path = f"{csv_main_path}/{file_name}.csv"
            df = pd.read_csv(csv_path)
            perf_data_dict.update(calculate_quantiles(df, col_name, prefix))
    except (FileNotFoundError, IndexError) as e:
        console.print_exception(show_locals=True)
        raise RuntimeError(f"Error processing benchmark output: {e}") from e

    return PerformanceData(**perf_data_dict)


def compare_metrics(
    prev_metrics: Dict[str, Dict[str, float]],
    new_metrics: Dict[str, Dict[str, float]],
    threshold_pct: int,
) -> Dict[str, Dict[str, PerformanceMetricComparison]]:
    """
    Compares two sets of performance metrics and identifies regressions.
    """
    comparison_results = {}
    for model, new_data in new_metrics.items():
        if model == "metadata":
            continue
        prev_data = PerformanceData(**prev_metrics.get(model, {}))
        new_data = PerformanceData(**new_data)
        comparison = {}
        for metric in new_data.get_metric_names():
            prev_value = getattr(prev_data, metric)
            new_value = getattr(new_data, metric)
            if prev_value is not None:
                diff_ms = new_value - prev_value
                regression_flag = diff_ms > (threshold_pct / 100) * prev_value
                comparison[metric] = PerformanceMetricComparison(
                    regression=regression_flag,
                    diff_ms=diff_ms,
                    diff_pct=100 * (diff_ms / prev_value),
                    prev_value=prev_value,
                    new_value=new_value,
                )
            else:
                comparison[metric] = PerformanceMetricComparison(
                    regression=None,
                    diff_ms=None,
                    diff_pct=None,
                    prev_value=None,
                    new_value=new_value,
                )
        comparison_results[model] = comparison
    return comparison_results


def save_comparison_results(
    comparison_output: Dict[str, any],
    output_dir: str,
    run_id: str,
    filename_prefix: str,
) -> None:
    """
    Saves the comparison results to a JSON file using utility function.
    """
    save_json_file(comparison_output, output_dir, f"{filename_prefix}_{run_id}.json")


def create_comparison_metadata(
    config: GitHubCIConfig, baseline_run_id: int | str | None
) -> Dict[str, any]:
    """
    Creates metadata for the comparison report.
    """
    metadata = PerformanceComparisonMetadata(
        regression_threshold_pct=config.regression_threshold_pct,
        date=datetime.now().isoformat(),
        workflow_run_id=config.run_id,
        baseline_run_id=baseline_run_id,
        commit_sha=config.github_sha,
    )
    return metadata.__dict__


def _update_json_file(
    json_path: str, key: str, new_data: PerformanceData, config: GitHubCIConfig
) -> None:
    """Updates or creates a JSON file with new performance data."""

    json_dir = os.path.dirname(json_path)
    os.makedirs(json_dir, exist_ok=True)

    try:
        with open(json_path, "r") as f:
            existing_data: Dict[str, any] = json.load(f)
    except (FileNotFoundError, json.JSONDecodeError):
        existing_data = {}

    existing_data[key] = new_data.to_dict()

    if (
        "metadata" not in existing_data
    ):  # minimal metadata just for run context if needed
        existing_data["metadata"] = PerformanceMetadata(
            date=datetime.now().isoformat(),
            workflow_run_id=config.run_id,
            commit_sha=config.github_sha,
        ).__dict__

    with open(json_path, "w") as f:
        json.dump(existing_data, f, indent=2, separators=(",", ": "))


def report_performance_comparison(
    comparison_results_data: Dict, regression_threshold_pct: int
) -> bool:
    """Displays performance comparison results in a formatted table.

    returns:
        bool: True if any regressions were detected, False otherwise.
    """

    console.rule("[bold magenta]Regression report start[/bold magenta]")

    perf_regression = False
    if comparison_results_data:
        for model_name, model_comparison in comparison_results_data.items():
            model_regressed = False
            table = Table(title=f"[bold blue]{model_name}[/bold blue]")
            table.add_column("Metric", style="cyan")
            table.add_column("Previous value", style="magenta")
            table.add_column("New value", style="green")
            table.add_column("Diff (ms)", style="yellow")
            table.add_column("Diff (%)", style="red")
            table.add_column("Regression", style="bold")

            for metric_name, metric_comparison in model_comparison.items():
                regression_status = (
                    "[bold red]YES[/bold red]"
                    if metric_comparison.regression
                    else "[green]NO[/green]"
                )

                diff_pct_style = (
                    "[green]"
                    if metric_comparison.diff_pct <= regression_threshold_pct
                    else "[red]"
                )
                table.add_row(
                    metric_name,
                    f"{metric_comparison.prev_value:.6f}",
                    f"{metric_comparison.new_value:.6f}",
                    f"{metric_comparison.diff_ms:.6f}",
                    f"{diff_pct_style}{metric_comparison.diff_pct:.2f}%{diff_pct_style}",
                    regression_status,
                )
                if metric_comparison.regression:
                    perf_regression = True
                    model_regressed = True

            if model_regressed:
                console.print(table)
                console.rule()

        if perf_regression:
            console.print("[bold red]Performance regression detected![/bold red]")
        else:
            console.print("[green]No performance regression detected.[/green]")
    return perf_regression
