import time
from collections import defaultdict
from dataclasses import dataclass
from queue import Empty, Queue
from threading import Event, Thread
from typing import Dict, List, Optional, Tuple

import zmq

from vajra.config import SystemConfig
from vajra.core.datatypes import (
    RequestOutput,
    SamplerOutputs,
    SchedulerOutput,
    Sequence,
    StepInputs,
    StepMicrobatchOuputs,
    StepOutputs,
)
from vajra.engine.base_llm_engine import BaseLLMEngine
from vajra.logger import init_logger
from vajra.utils.threading_utils import exit_on_error, synchronized

logger = init_logger(__name__)

SCHEDULER_LOOP_DELAY = 0.01


@dataclass(frozen=True)
class ScheduleStageOutputs:
    ignored_seqs: List[Sequence]
    seqs: List[Sequence]
    scheduler_output: SchedulerOutput
    start_time: float


class PipelineParallelLLMEngine(BaseLLMEngine):
    """An LLM engine that receives requests and generates texts.

    This is the main class for the Vajra engine. It receives requests
    from clients and generates texts from the LLM. It includes a tokenizer, a
    language model (possibly distributed across multiple GPUs), and GPU memory
    space allocated for intermediate states (aka KV cache). This class utilizes
    iteration-level scheduling and efficient memory management to maximize the
    serving throughput.

    Args:
        config; System Config: The system configuration for the engine.
    """

    def __init__(
        self,
        config: SystemConfig,
    ) -> None:
        super().__init__(config)

        # Create the request queue.
        self.has_started_execution_loops = False
        self.scheduler_output_queue = Queue()
        self.output_queue = Queue()
        self.schedule_event = Event()
        self.microbatch_watch_queue = Queue()
        self.schedule_thread = Thread(target=self._schedule_loop, daemon=True)
        self.microbatch_watch_thread = Thread(
            target=self._microbatch_watch_loop, daemon=True
        )
        self.output_thread = Thread(target=self._output_loop, daemon=True)
        self.scheduler_timer_thread = Thread(
            target=self._scheduler_timer_loop, daemon=True
        )
        self.process_output_queue = Queue()

        self.pending_step_outputs: List[Tuple[SchedulerOutput, SamplerOutputs]] = []

    def _init_zmq_sockets(self):
        super()._init_zmq_sockets()
        # PULL socket for microbatch completion signal
        self.microbatch_socket = self.zmq_context.socket(zmq.PULL)
        self._bind_zmq_socket(
            self.microbatch_socket, self.comm_info.microbatch_socket_port
        )

    def _validate_parallel_config(self) -> None:
        assert self.config.parallel_config.pipeline_parallel_size > 1

    def start_execution_loops(self) -> None:
        """Starts the execution loop."""
        self.has_started_execution_loops = True
        self.schedule_event.set()
        self.schedule_thread.start()
        self.output_thread.start()
        self.scheduler_timer_thread.start()
        self.microbatch_watch_thread.start()

    def add_seq_to_scheduler(self, seq: Sequence) -> None:
        self.scheduler.add_seq(seq)
        self.schedule_event.set()

    @exit_on_error
    def _scheduler_timer_loop(self) -> None:
        while True:
            time.sleep(SCHEDULER_LOOP_DELAY)
            self.schedule_event.set()

    def _get_worker_impl(self):
        # Lazy import the Worker to avoid importing torch.cuda/xformers
        # before CUDA_VISIBLE_DEVICES is set in the Worker
        from vajra.worker.pipeline_parallel_worker import PipelineParallelWorker

        return PipelineParallelWorker

    @synchronized
    def _append_pending_step_output(
        self, scheduler_output: SchedulerOutput, sampler_outputs: SamplerOutputs
    ) -> None:
        self.pending_step_outputs.append((scheduler_output, sampler_outputs))

    @synchronized
    def _get_pending_step_outputs(
        self,
    ) -> List[Tuple[SchedulerOutput, SamplerOutputs]]:
        pending_step_outputs = self.pending_step_outputs
        self.pending_step_outputs = []
        return pending_step_outputs

    @exit_on_error
    def _schedule_loop(self) -> None:
        while True:
            self.schedule_event.wait()
            self.schedule_event.clear()

            start_time = time.time()

            scheduler_output = self.scheduler.schedule()

            if scheduler_output.has_no_output:
                continue

            ignored_seqs, seqs = self.seq_manager.on_schedule(scheduler_output)

            self.scheduler_output_queue.put(
                ScheduleStageOutputs(
                    ignored_seqs,
                    seqs,
                    scheduler_output,
                    start_time,
                )
            )

            end_time = time.time()

            if not scheduler_output.is_empty:
                self.microbatch_watch_queue.put(scheduler_output)
                self.enqueue_socket.send_pyobj(
                    StepInputs(
                        scheduler_output,
                        new_seqs=self._get_new_seqs(),
                        pending_step_outputs=self._get_pending_step_outputs(),
                    )
                )

            self.metrics_store.on_schedule(
                scheduler_output.seq_schedule_metadata_list, start_time, end_time
            )

    @exit_on_error
    def _microbatch_watch_loop(self) -> None:
        pending_microbatch_outputs: Dict[int, List[StepMicrobatchOuputs]] = defaultdict(
            list
        )

        while True:
            scheduler_output = self.microbatch_watch_queue.get()
            schedule_id = scheduler_output.id
            num_microbatch_outputs_received = 0

            num_microbatch_outputs_received += len(
                pending_microbatch_outputs[schedule_id]
            )
            del pending_microbatch_outputs[schedule_id]

            while (
                num_microbatch_outputs_received
                < self.config.parallel_config.kv_parallel_size
            ):
                step_microbatch_outputs = self.microbatch_socket.recv_pyobj()
                if step_microbatch_outputs.schedule_id != schedule_id:
                    pending_microbatch_outputs[
                        step_microbatch_outputs.schedule_id
                    ].append(step_microbatch_outputs)
                    continue

                num_microbatch_outputs_received += 1

            self.seq_manager.on_stage_completed(scheduler_output)
            self.scheduler.on_stage_completed(
                [
                    self.seq_manager.get_seq(s.seq_id)
                    for s in scheduler_output.seq_schedule_metadata_list
                ]
            )
            self.process_output_queue.put(None)
            self.schedule_event.set()

    @exit_on_error
    def _output_loop(self) -> None:
        pending_step_outputs: Dict[int, List[StepOutputs]] = defaultdict(list)

        while True:
            scheduler_stage_output = self.scheduler_output_queue.get()
            schedule_id = scheduler_stage_output.scheduler_output.id
            num_step_outputs_received = 0

            all_sampler_outputs: List[Optional[SamplerOutputs]] = []

            all_sampler_outputs.extend(pending_step_outputs[schedule_id])
            num_step_outputs_received += len(pending_step_outputs[schedule_id])
            del pending_step_outputs[schedule_id]

            while (
                num_step_outputs_received < self.config.parallel_config.kv_parallel_size
            ):
                step_output = self.output_socket.recv_pyobj()
                if step_output.schedule_id != schedule_id:
                    pending_step_outputs[step_output.schedule_id].append(
                        step_output.sampler_outputs
                    )
                    continue
                all_sampler_outputs.append(step_output.sampler_outputs)
                num_step_outputs_received += 1

            combined_sampler_outputs = self._combine_sampler_outputs(
                all_sampler_outputs, scheduler_stage_output.seqs
            )

            self._append_pending_step_output(
                scheduler_stage_output.scheduler_output, combined_sampler_outputs
            )

            self.process_output_queue.get()
            all_request_outputs = self._on_step_completed(
                scheduler_stage_output.scheduler_output,
                scheduler_stage_output.ignored_seqs,
                scheduler_stage_output.seqs,
                combined_sampler_outputs,
                scheduler_stage_output.start_time,
            )
            self.schedule_event.set()
            self.output_queue.put(all_request_outputs)

    def step(self, block: bool = True) -> List[RequestOutput]:
        """Performs one decoding iteration and returns newly generated results.

        This function performs one decoding iteration of the engine.
        This version does everything asynchronously and returns the results
        """
        if not self.has_started_execution_loops:
            self.start_execution_loops()

        if block:
            return self.output_queue.get()

        try:
            return self.output_queue.get(block=False)
        except Empty:
            return []
